/*
 * Copyright (c) 2014, Ericsson AB. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, this
 * list of conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 */

/*
 * Openh264 GStreamer plugin
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gst/gst.h>
#include <wels/codec_api.h>
#include <wels/codec_ver.h>
#include <string.h>
#include "gstopenh264dec.h"
#include "gstopenh264enc.h"

static gboolean
plugin_init (GstPlugin * plugin)
{
  /* g_stCodecVersion is the version detected at build time as defined in the
   * headers and WelsGetCodecVersion() is the version detected at runtime.
   * This is a safeguard to avoid crashes since OpenH264  has been changing
   * ABI without changing the SONAME.
   */
  OpenH264Version libver = WelsGetCodecVersion ();
  if (memcmp (&libver, &g_stCodecVersion, sizeof (libver)) == 0) {
    gst_element_register (plugin, "openh264dec", GST_RANK_MARGINAL,
        GST_TYPE_OPENH264DEC);
    gst_element_register (plugin, "openh264enc", GST_RANK_MARGINAL,
        GST_TYPE_OPENH264ENC);
  } else {
    GST_ERROR ("Incorrect library version loaded, expecting %s", g_strCodecVer);
  }
  return TRUE;
}

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    openh264,
    "OpenH264 encoder/decoder plugin",
    plugin_init, VERSION, "BSD", GST_PACKAGE_NAME, GST_PACKAGE_ORIGIN)
